#include "../slim.h"
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "../common/objects.h"
#include "../common/scene.h"
#include "./obj_parser.h"
#include "./slm_parser.h"
#include "../common/debug.h"
#include "../shader/shaders.h"
#include "../intersect/sphere.h"
#include "../intersect/cylinder.h"
#include "../intersect/plane.h"
#include "../intersect/disc.h"
#include "../intersect/rectangle.h"
#include "../shader/lighting.h"
#include "parser.h"
#include "../load/setup_scene.h"
#include "../simulation/simulation.h"

extern int current_sim;
extern int sim_type;
extern int DEBUG_LEVEL;
extern color red;
extern color white;

//move this to scene object file later
void setup_generic_scene(scene_data *scene)
{
	//light
	scene->lights[0] = make_object(LIGHT_INFINITE, 0);
	set_tri(&scene->lights[0]->pos, 1, 0.75, 0.34);
	normalize(&scene->lights[0]->pos);
	make_color(&scene->lights[0]->amb, 255, 255, 255);
	make_color(&scene->lights[0]->spec, 255, 255, 255);
	make_color(&scene->lights[0]->diff, 255, 255, 255);
	printd(MESSAGE, "Light[%i], type[%i-light] - ok\n",	0, scene->lights[0]->obj_type);
	
	//camera
	set_tri(&scene->camera->lookP, 0, 0, 0);
	set_tri(&scene->camera->look, 0, 0, 0);
	set_tri(&scene->camera->pos, 20, 20, 20);
	set_tri(&scene->camera->rot, 0, 0, 0);
	set_tri(&scene->camera->up, 0, 0, 1);
	printd(MESSAGE, "setting up camera\n");
	scene->camera->units = 5;

	//misc
	scene->height = 401;
	scene->width = 401;
	scene->max_recurs = 3;
	scene->sub_dis = 16;
	scene->antialiasing = 0;
	scene->sub_var = 10;
	scene->sub_level = 1;
}


int parse_obj_scene(scene_data* scene)
{
	void* shader_ptr;
	char *line_pos;
	char *col_pos;
	list vertex_list;
	
	int linenumber = 0;
	
	//TODO: fix this crap
	//TERRIBLE HACK
	//allows the background shader to be set while parsing scene file
	setup_background_object();
	
	char *current_line;
	char *first_token;
	
	list_make(&vertex_list, 10, 1);

	current_line = strtok_r( (char *) scene->file_content, "\n\r", &line_pos);

	/* while the string is not null parse it */
	do
	{
		first_token = strtok_r(current_line, " \t\n\r", &col_pos);
		shader_ptr = (void*) &general_shader;

		//skip comments
		if( strncmp("//", first_token, 2)==0 || strncmp("#", first_token, 1)==0)
			continue;
		else if( strncmp("/*", first_token, 2) == 0)
			while( strncmp("*/", strtok_r(NULL, " \t\n\r", &line_pos), 2)!=0 )
				linenumber++;

		/* process objects */
		else if( first_token[0] == 'v' ) //process vertex
		{
			vector *v = (vector*)malloc(sizeof(vector));
			
			v->x = atof( strtok_r(NULL, " \t", &col_pos));
			v->y = atof( strtok_r(NULL, " \t", &col_pos));
			v->z = atof( strtok_r(NULL, " \t", &col_pos));
			
			list_add_item(&vertex_list, v, NULL);
		}
		else if( first_token[0] == 'f' ) //process face
		{
			int indices[4];
			int i;
			vector *vertices[4];
			object *rect;

			for(i=0; i<4; i++)
			{
				indices[i] = atoi( strtok_r(NULL, " \t", &col_pos));
				vertices[i] = (vector*) list_get_index(&vertex_list, indices[i]-1);
			}
				
			rect = rectangle_create(main_scene->num_objects);
			rectangle_from_points(rect, vertices);

			make_color(&rect->amb, 250, 250, 150);
			make_color(&rect->diff, 250, 250, 150);
			make_color(&rect->spec, 250, 250, 250);
			rect->shader = (void* (*)(void*, color*))shader_ptr;
			
			set_misc(rect, 0, 0, 0, 0, 100);
			
			scene_add_object(rect, main_scene);
		}
		else if( first_token[0] == 'p' ) //process point
		{
			//make a small sphere to represent the point
			vector *v;
			int i;
			
			i = atoi( strtok_r(NULL, " \t", &col_pos));
			v = (vector*) list_get_index(&vertex_list, i-1);
			

			object *obj = sphere_create();
			copy_vector(v, &obj->pos);
			
			//add sphere to scene models
			scene_add_object(obj, main_scene);
			
			obj->radius = 2 + 0.1 * (double)rand()/RAND_MAX;
			
			color_copy(&obj->diff, &red);
			color_copy(&obj->amb, &red);
			//	color_copy(&obj->amb, &black);
			color_from_number(&obj->diff, obj->id);
			color_from_number(&obj->amb, obj->id);
			color_copy(&obj->spec, &white);
		}
		else if( first_token[0] == 'o' ) //object name
		{ }
		else if( first_token[0] == 's' ) //smoothing
		{ }
		else if( first_token[0] == 'u' ) //use material
		{ }
		

		else
		{
			printd(ERROR, "Error in scene code at line %i:\n\t%s\n",
					linenumber,  current_line);
			return 0;
		}
	}
	while( current_line != NULL && (current_line = strtok_r(NULL, "\n\r", &line_pos)) != NULL );
	
	list_delete_all(&vertex_list);
	setup_generic_scene(main_scene);
	return 1;
}

